/*
 * @author     Ekawat (ice) Homsirikamol
 * @version    1.0
 * @copyright  Copyright (c) 2017 Cryptographic Engineering Research Group
 *             ECE Department, George Mason University Fairfax, VA, U.S.A.
 *             All rights Reserved.
 * @license    This project is released under the GNU Public License.
 *             The license and distribution terms for this file may be
 *             found in the file LICENSE in this distribution or at
 *             http://www.gnu.org/licenses/gpl-3.0.txt
 * @note       This implementation is derived in part from the reference implementation
 *             included in SUPERCOP-20171020, with the list of co-authors listed in ref/designers
 *             
 *             This is publicly available encryption source code that falls
 *             under the License Exception TSU (Technology and software-
 *             unrestricted)
 */

#if !__SYNTHESIS__
#include <iostream>
#endif

#include "ap_int.h"
#include "hls_stream.h"
#include "hls_interface.h"
#include "hls_aes.h"
#include "hls_utils.h"
#include <string.h>
using namespace hls;


enum fsms_t {S_INIT,
    S_DATA,
    S_LEN, S_FINAL, S_OUT};

#define INIT_ROUND 15
#define FINAL_ROUND 20
#define WEIRD_BUG 0

/* -------------------------------------------------------*/

#if !__SYNTHESIS__
    #define DBG 0
#endif  // if !__SYNTHESIS__
#define SYNTHESIS_FAIL 0


#if !__SYNTHESIS__
ap_uint<128> reshape16(aes_t in) {
    size_t i;
    ap_uint<128> out = 0;
    for (i=0; i<16; i++) {
        out ^= ((ap_uint<128>)in[i]) << (15-i)*8;
    }
    return out;
}
void printState(aes_t t3[3], aes_t t4[4], aes_t t6[6]) {
    size_t i;
    std::cout << "T3: ";
    for (i=0; i<3; i++) {
        std::cout << std::hex << std::setw(16) << reshape16(t3[i]) << " ";
    }
    std::cout << std::endl << "T4: ";
    for (i=0; i<4; i++) {
        std::cout << std::hex << std::setw(16) << reshape16(t4[i]) << " ";
    }
    std::cout << std::endl << "T6: ";
    for (i=0; i<6; i++) {
        std::cout << std::hex << std::setw(16) << reshape16(t6[i]) << " ";
    }
    std::cout << std::endl;
}
#endif

/* -------------------------------------------------------*/

void hls_ciphercore (
    BDO_STREAM& bdo_bus,
    bool *msg_auth_valid,
    BDI_STREAM& bdi_bus,
    KEY_STREAM& key_bus,
    bool decrypt,
    volatile bool key_update
    )
{
    /* Top-Level Interface */
#pragma HLS DATA_PACK variable=bdo_bus
#pragma HLS DATA_PACK variable=bdi_bus
#pragma HLS DATA_PACK variable=key_bus
#pragma HLS INTERFACE port=bdo_bus axis
#pragma HLS INTERFACE port=bdi_bus axis
#pragma HLS INTERFACE port=key_bus axis
#pragma HLS INTERFACE ap_vld  port=msg_auth_valid
#if __SYNTHESIS__
#else
    size_t dbg_loop = 0;
#endif

    size_t i;
    public_bus bdi;
    secret_bus key;


    bool loop = 1;
    fsms_t fsm = S_INIT;
    
    static aes_t reg_key;
    const static aes_t Z0 = {0x42,0x8a,0x2f,0x98,0xd7,0x28,0xae,0x22,0x71,0x37,0x44,0x91,0x23,0xef,0x65,0xcd};
    const static aes_t Z1 = {0xb5,0xc0,0xfb,0xcf,0xec,0x4d,0x3b,0x2f,0xe9,0xb5,0xdb,0xa5,0x81,0x89,0xdb,0xbc};
    
    
    array_t inputInit;
    aes_t reg_t3[3];
    aes_t reg_t4[4];
    aes_t reg_t6[6];
    bool is_eoi;
    ap_uint<32> len_a = 0;
    ap_uint<32> len_d = 0;
    ap_uint<5> cnt = 0;
    
#pragma HLS ARRAY_RESHAPE variable=reg_key   complete dim=1
#pragma HLS ARRAY_RESHAPE variable=Z0   complete dim=1
#pragma HLS ARRAY_RESHAPE variable=Z1   complete dim=1
#pragma HLS ARRAY_RESHAPE variable=inputInit  complete dim=1
#pragma HLS ARRAY_RESHAPE variable=reg_t3    complete dim=1
#pragma HLS ARRAY_RESHAPE variable=reg_t3[0] complete dim=1
#pragma HLS ARRAY_RESHAPE variable=reg_t3[1] complete dim=1
#pragma HLS ARRAY_RESHAPE variable=reg_t3[2] complete dim=1
#pragma HLS ARRAY_RESHAPE variable=reg_t4    complete dim=1
#pragma HLS ARRAY_RESHAPE variable=reg_t4[0] complete dim=1
#pragma HLS ARRAY_RESHAPE variable=reg_t4[1] complete dim=1
#pragma HLS ARRAY_RESHAPE variable=reg_t4[2] complete dim=1
#pragma HLS ARRAY_RESHAPE variable=reg_t4[3] complete dim=1
#pragma HLS ARRAY_RESHAPE variable=reg_t6    complete dim=1
#pragma HLS ARRAY_RESHAPE variable=reg_t6[0] complete dim=1
#pragma HLS ARRAY_RESHAPE variable=reg_t6[1] complete dim=1
#pragma HLS ARRAY_RESHAPE variable=reg_t6[2] complete dim=1
#pragma HLS ARRAY_RESHAPE variable=reg_t6[3] complete dim=1
#pragma HLS ARRAY_RESHAPE variable=reg_t6[4] complete dim=1
#pragma HLS ARRAY_RESHAPE variable=reg_t6[5] complete dim=1


    // Key initialization (if needed)
    if (key_update == 1) {
        LOAD_KEY: for (i=0; i<4; i++) {
            key_bus.read(key);
            reg_key[i*4+0] = (key.data >> 24) & 0xFF;
            reg_key[i*4+1] = (key.data >> 16) & 0xFF;
            reg_key[i*4+2] = (key.data >>  8) & 0xFF;
            reg_key[i*4+3] = (key.data >>  0) & 0xFF;
        }
    }

    // Initialize here
    bdi_bus.read(bdi); reshape_block(inputInit, bdi.data);
    is_eoi = bdi.eoi;
    
    Copy16(reg_t3[0], reg_key);
    Copy16(reg_t3[1], reg_key);
    Copy16(reg_t3[2], inputInit);
    
    Copy16(reg_t4[0], reg_key);
    Copy16(reg_t4[1], reg_key);
    Copy16(reg_t4[2], inputInit);
    Copy16(reg_t4[3], Z0);
    
    Copy16(reg_t6[0], reg_key);
    Copy16(reg_t6[1], reg_key);
    Copy16(reg_t6[2], inputInit);
    Copy16(reg_t6[3], Z1);
    Clear(reg_t6[4], 16);
    Clear(reg_t6[5], 16);
    
MAIN_LOOP: while (loop) {
        // Add this clause to prevent infinite loop during debugging
#if !__SYNTHESIS__
        if (dbg_loop > 100) return;
        else dbg_loop++;
#endif
#if DBG
        std::cout << "Current fsm = " << fsm << std::endl;        
#endif
        // Default values to make it combinational
        bool eoi = 0;
        bool is_data = 0;
        data_bytes_t vbytes = 0;
        ap_uint<6> size = 0;
        aes_t t3_sel;
        aes_t t4_sel;
        aes_t t6_sel;
        //aes_t t3[2];
        aes_t t3_0, t3_1;
        aes_t t4_0, t4_1;
        aes_t t6_0, t6_1;
        aes_t t3_rkey;
        aes_t t4_rkey;
        aes_t t6_rkey;
        array_t input;
#pragma HLS ARRAY_RESHAPE variable=input  complete dim=1
#pragma HLS ARRAY_RESHAPE variable=t3_sel  complete dim=1
#pragma HLS ARRAY_RESHAPE variable=t4_sel  complete dim=1
#pragma HLS ARRAY_RESHAPE variable=t6_sel  complete dim=1
#pragma HLS ARRAY_RESHAPE variable=t3_rkey   complete dim=1
#pragma HLS ARRAY_RESHAPE variable=t4_rkey   complete dim=1
#pragma HLS ARRAY_RESHAPE variable=t6_rkey   complete dim=1
        
        
        // Pre-processing
#if 0
        switch (fsm) {
        case S_INIT:            
            Copy16(t3_sel, Z0);
            Copy16(t4_sel, Z1);
            Copy16(t6_sel, Z0);
            break;
        case S_DATA:
            #if DBG
                std::cout << "Reading data ..." << std::endl;
            #endif
            bdi_bus.read(bdi); 
            // Update temporary variables
            reshape_block(input, bdi.data);
            eoi = bdi.eoi;
            vbytes = bdi.valid_bytes;
            size = bdi.size;
            is_data = (bdi.type == TYPE_D) ? 1:0;
            if (is_data && decrypt) {
                Clear(t3_sel, 16);
                Clear(t4_sel, 16);
                Clear(t6_sel, 16);
            } else {
                Copy16(t3_sel, input);
                Copy16(t4_sel, input+16);
                Xor16(t6_sel, input, input+16);
            }
            
            break;
        case S_LEN:
            Clear(t3_sel, 12);
            Clear(t4_sel, 12);
            Clear(t6_sel, 12);
            for (i=0; i<4; i++) {
			#pragma HLS UNROLL
                ap_uint<8> len_a8 = (len_a >> (8*(3-i))) & 0xFF;
                ap_uint<8> len_d8 = (len_d >> (8*(3-i))) & 0xFF;
                t3_sel[12+i] = len_a8;
                t4_sel[12+i] = len_d8;
                t6_sel[12+i] = len_a8 ^ len_d8;
            }
            break;
        case S_FINAL:
            Copy16(t3_sel, Z1);
            Copy16(t4_sel, Z0);
            Copy16(t6_sel, Z1);
            break;
        default: // S_OUT
            Clear(t3_sel, 16);
            Clear(t4_sel, 16);
            Clear(t6_sel, 16);
        }
#else
        if (S_DATA == fsm) {
            bdi_bus.read(bdi);
            reshape_block(input, bdi.data);

			is_data = (bdi.type == TYPE_D) ? 1:0;
			eoi = bdi.eoi;
			size = bdi.size;
			vbytes = bdi.valid_bytes;
			#if DBG
			 std::cout << "Reading ..." << std::endl;
			#endif
        } else {
			clear(input);
        }


        if (S_INIT == fsm) {
            Copy16(t3_sel, Z0);
            Copy16(t4_sel, Z1);
            Copy16(t6_sel, Z0);
        } else if (S_DATA == fsm) {
            if (is_data && decrypt) {
                Clear(t3_sel, 16);
                Clear(t4_sel, 16);
                Clear(t6_sel, 16);
            } else {
                Copy16(t3_sel, input);
                Copy16(t4_sel, input+16);
                Xor16(t6_sel, input, input+16);
            }
        } else if (S_LEN == fsm) {
            Clear(t3_sel, 12);
            Clear(t4_sel, 12);
            Clear(t6_sel, 12);
            for (i=0; i<4; i++) {
			#pragma HLS UNROLL
                ap_uint<8> len_a8 = (len_a >> (8*(3-i))) & 0xFF;
                ap_uint<8> len_d8 = (len_d >> (8*(3-i))) & 0xFF;
                t3_sel[12+i] = len_a8;
                t4_sel[12+i] = len_d8;
                t6_sel[12+i] = len_a8 ^ len_d8;
            }
        } else if (S_FINAL == fsm) {
            Copy16(t3_sel, Z1);
            Copy16(t4_sel, Z0);
            Copy16(t6_sel, Z1);
        } else {
            Clear(t3_sel, 16);
            Clear(t4_sel, 16);
            Clear(t6_sel, 16);
        }
#endif
        
        Xor16(t3_rkey, reg_t3[0], t3_sel);
        Xor16(t4_rkey, reg_t4[0], t4_sel);
        Xor16(t6_rkey, reg_t6[0], t6_sel);

        /* =====*/
        /* Core */
        /* =====*/
#if DBG
        std::cout << "Count: " << std::dec << cnt << std::endl;
        std::cout << "Input: " << std::endl;
        printState(reg_t3, reg_t4, reg_t6);
        std::cout << "t3_sel: " << std::hex << std::setw(16) << reshape16(t3_sel) << std::endl;
        std::cout << "t4_sel: " << std::hex << std::setw(16) << reshape16(t4_sel) << std::endl;
        std::cout << "t6_sel: " << std::hex << std::setw(16) << reshape16(t6_sel) << std::endl;
        std::cout << "t3_rkey: " << std::hex << std::setw(16) << reshape16(t3_rkey) << std::endl;
        std::cout << "t4_rkey: " << std::hex << std::setw(16) << reshape16(t4_rkey) << std::endl;
        std::cout << "t6_rkey: " << std::hex << std::setw(16) << reshape16(t6_rkey) << std::endl;
#endif
        AESRoundBasic(t3_0, reg_t3[2], t3_rkey);
        AESRoundBasic(t3_1, reg_t3[0], Z0);
        
        AESRoundBasic(t4_0, reg_t4[3], t4_rkey);
        AESRoundBasic(t4_1, reg_t4[0], Z0);
        
        AESRoundBasic(t6_0, reg_t6[5], t6_rkey);
        AESRoundBasic(t6_1, reg_t6[0], Z0);
#if DBG
        std::cout << "\tOutput: " << std::endl;
        printState(t3, t4, t6);
#endif

        /* ========= */
        /* Post-processing  */
        /* ========= */
        aes_t c0, c1, m0, m1;
        array_t output;
        
        // Output   
        OLOOP: 
        for (i=0; i<16; i++) {
        #pragma HLS UNROLL
            c0[i] = t3_0[i] ^ reg_t3[2-1][i] ^ t4_1[i] ^ (reg_t6[3-1][i] & reg_t4[3-1][i]);
            c1[i] = t6_0[i] ^ reg_t4[2-1][i] ^ t3_1[i] ^ (reg_t6[5-1][i] & reg_t3[2-1][i]);
            m0[i] = input[i] ^ c0[i];
            m1[i] = input[i+16] ^ c1[i] ^ m0[i];
        }

        if (decrypt) {
            Copy16(output, m0);
            Copy16(output+16, m1);                
        } else {
            Copy16(output, c0);
            Copy16(output+16, c1);
        }
		#if DBG
			std::cout << "Writing output ..." << std::endl;
		#endif

        // Register update                
        Copy16(reg_t3[2], reg_t3[2-1]);
        Copy16(reg_t3[1], t3_1);

        Copy16(reg_t4[3], reg_t4[3-1]);
        Copy16(reg_t4[2], reg_t4[2-1]);
        Copy16(reg_t4[1], t4_1);

        Copy16(reg_t6[5], reg_t6[5-1]);
        Copy16(reg_t6[4], reg_t6[4-1]);
        Copy16(reg_t6[3], reg_t6[3-1]);
        Copy16(reg_t6[2], reg_t6[2-1]);
        Copy16(reg_t6[1], t6_1);
        if (is_data && decrypt) {
#if WEIRD_BUG
            array_t inputTmp;
            array_t msg;
#else
            ap_uint<8> inputTmp[32];
            ap_uint<8> msg[32];
#endif
            aes_t msg01;

#if WEIRD_BUG
            std::cout << "[BUG] Before Copying .. " << std::endl;
            std::cout << "output: " << std::hex << std::setw(16) << reshape16(output) << " " << reshape16(output+16) << std::endl;
#endif
            Copy16(inputTmp, m0);
            Copy16(inputTmp+16, m1);
            byte_valid(msg, inputTmp, vbytes);
#if WEIRD_BUG
            std::cout << "[BUG] After Copying .. " << std::endl;
            std::cout << "output: " << std::hex << std::setw(16) << reshape16(output) << " " << reshape16(output+16) << std::endl;
#endif

            Xor16(reg_t3[0], t3_0, msg);
            Xor16(reg_t4[0], t4_0, msg+16);
            Xor16(msg01, msg, msg+16);
            Xor16(reg_t6[0], t6_0, msg01);
        } else {
            Copy16(reg_t3[0], t3_0);
            Copy16(reg_t4[0], t4_0);
            Copy16(reg_t6[0], t6_0);
        }



        // State transition
        switch (fsm) {
        case S_INIT:
            if (cnt == INIT_ROUND-1) {
                cnt = 0;
                if (is_eoi) fsm = S_LEN;
                else fsm = S_DATA;
            } else {
                cnt++;
            }            
            break;
        case S_DATA:
            if (is_data) len_d += size;
            else len_a += size;
            if (eoi) fsm = S_LEN;
            if (is_data) {
                output_bus bdo;
                bdo.data = reshape_array(output);
                bdo_bus.write(bdo);
            }
            break;
        case S_LEN:
            fsm = S_FINAL;
            break;
        default: // S_FINAL
            if (cnt == FINAL_ROUND-1) {
                loop = 0;
            } else {
                cnt++;
            }
        }
    }
#if !__SYNTHESIS__
    std::cout << "Main loop exited. Program terminated." << std::endl;
#endif
    array_t tag_output;
    data_t tagBlock;

    for (i=0; i<16; i++) {
	#pragma HLS UNROLL
    	tag_output[i] = reg_t3[0][i] ^ reg_t3[1][i] ^ reg_t3[2][i] ^ \
                 reg_t4[0][i] ^ reg_t4[1][i] ^ reg_t4[2][i] ^ reg_t4[3][i] ^ \
                 reg_t6[0][i] ^ reg_t6[1][i] ^ reg_t6[2][i] ^ reg_t6[3][i] ^ reg_t6[4][i] ^ reg_t6[5][i];
    }
	Clear(tag_output+16, 16);

    tagBlock = reshape_array(tag_output);
	if (decrypt) {
		#if DBG
			std::cout << "Reading tag ..." << std::endl;
		#endif
		bdi_bus.read(bdi);
		*msg_auth_valid = (bdi.data == tagBlock) ? 1:0;
	} else {
		#if DBG
			std::cout << "Writing tag ..." << std::endl;
		#endif
        output_bus bdo;
		bdo.data = reshape_array(tag_output);
		bdo_bus.write(bdo);
	}
	loop = 0;
    return;
}
